"use strict";Object.defineProperty(exports,"__esModule",{value:true});const http_1=require("./instrumentations/http");const logger_1=require("./logger");const metrics_1=require("./metrics");const transaction_1=require("./transaction");const wrapper_1=require("./wrapper");const config_1=require("./config");const async_1=require("./async");const dripstat_server_1=require("./dripstat_server");const instr_loader_1=require("./instrumentations/discovery/instr_loader");const trace_data_1=require("./trace_data");const trace_sampler_1=require("./trace_sampler");const ports_1=require("./ports");const TRANSACTION_CONTEXT_KEY="__agent__";const TRANSACTION_KEY="___transaction___";const EXCEPTION_LIST_MAX=20;class AgentData{constructor(){this._transactionStats={};this._metrics={};this._errors={};this._traces={};this.exceptionDataCounter=0}get transactionStats(){return this._transactionStats}get metrics(){return this._metrics}get errors(){return this._errors}get traces(){return this._traces}collectError(transaction){if(!transaction.stacktrace){return}if(!transaction.finished){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug(`Cannot add an error from an unfinished transaction`);return}const transName=transaction.getName();const parsed=trace_data_1.ExceptionParser.parse(transaction.stacktrace);const key=`${transName} ${parsed.exceptionCls}`;if(this._errors[key]){const existing=this._errors[key];existing.count++;return}if(this.exceptionDataCounter<=EXCEPTION_LIST_MAX){this._errors[key]=new trace_data_1.ExceptionData(parsed,transName,{});this.exceptionDataCounter++}}addMetrics(transaction){this.collectTrace(transaction);this.collectError(transaction);const transactionName=transaction.getName();let metricsForThisTrans=this._metrics[transactionName];if(!metricsForThisTrans){metricsForThisTrans={};this._metrics[transactionName]=metricsForThisTrans}if(!this._transactionStats[transactionName]){this._transactionStats[transactionName]=new metrics_1.Stats}this._transactionStats[transactionName].addTime(transaction.timer.time);for(const metricName in transaction.timers){let stats=metricsForThisTrans[metricName];if(!stats){stats=new metrics_1.Stats;metricsForThisTrans[metricName]=stats}const timers=transaction.timers[metricName];for(const timer of timers){if(timer.isStopped()){stats.addTime(timer.time);if(timer.isError){stats.addErrors(1)}}else{if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug(`Unstopped timer for ${metricName}`)}}}}collectTrace(transaction){if(!transaction.finished){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug(`Cannot add trace from a unfinished transaction`);return}if(transaction.timer.time<transaction.config.traceThreshold){return}const transName=transaction.getName();const parsed=trace_data_1.ExceptionParser.parse(transaction.stacktrace);const traceData=new trace_data_1.TraceData(transaction,new trace_data_1.ExceptionData(parsed,transaction.getName(),{}));const key=`${transName} ${parsed.exceptionCls}`;if(!this._traces[key]){this._traces[key]=new trace_sampler_1.TraceSampler}this._traces[key].recordTrace(traceData)}}exports.AgentData=AgentData;class Agent{constructor(config){this.config=config;this.instrumentations=[new http_1.HttpInstrumentation];this.discoverInstrumentations=instr_loader_1.dicoverInstrumentations;this.initiatedInstrumentations=new Set;this._data=new AgentData;this.dripStatServer=new dripstat_server_1.DripstatServer(this.config)}get data(){return this._data}initialize(){const errMsg=config_1.AgentConfigManager.verify(this.config);if(errMsg){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.ERROR))logger_1.agentLogger.error(`Disabling Agent due to Configuration Error : ${errMsg}`);return}for(const instr of this.instrumentations){instr.initialize(null)}this.patchImporter();this.doOnNextFullMinute()}doOnNextFullMinute(){const d=new Date;const millisToFullMinute=6e4-(d.getSeconds()*1e3+d.getMilliseconds());setTimeout(this.minuteTicks.bind(this),millisToFullMinute)}newWebTransaction(req,resp){const ctx={};const asyncId=async_1.storeAsyncId(ctx);const transaction=new transaction_1.WebTransaction(asyncId,this.config);transaction.initialize(req);ctx[TRANSACTION_CONTEXT_KEY]=transaction;req[TRANSACTION_KEY]=transaction;if(resp)resp[TRANSACTION_KEY]=transaction;return transaction}getExistingWebTransaction(reqOrResp){if(reqOrResp){const trans=reqOrResp[TRANSACTION_KEY];if(trans){return trans}}const asyncCtx=async_1.getAsyncContext();if(!asyncCtx){return null}return asyncCtx[TRANSACTION_CONTEXT_KEY]}getExistingWebTransactionAndStartMetric(metricName,type){const trans=this.getExistingWebTransaction();if(trans)trans.startTimer(metricName,type);return trans}finish(transaction){async_1.clearAsyncId(transaction.asyncId);if(transaction.finished){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug("Transaction already finished")}transaction.finished=true;const transTimer=transaction.timer;transTimer.isError=transaction.isError;transTimer.stop();if(transTimer.time>this.config.stackThreshold){transaction.recordRootStacktrace(new Error)}this._data.addMetrics(transaction)}patchImporter(){const that=this;const Module=require("module");let resolvedName=null;wrapper_1.wrapMethod(Module,"Module",["_resolveFilename"],function(resolve){return function(){return resolvedName=resolve.apply(this,arguments)}});wrapper_1.wrapMethod(Module,"Module",["_load"],function(load){return function(file){const res=load.apply(this,arguments);that.onModuleLoad(res,file,resolvedName);return res}})}onModuleLoad(nodule,name,resolvedName){this.tryInitializeInstrumentation(name,this.discoverInstrumentations[name],nodule)}tryInitializeInstrumentation(moduleName,instr,nodule){if(this.initiatedInstrumentations.has(moduleName)){return}const that=this;if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug("trying to initialize instrumentation %s",moduleName);if(!instr){return}if(this.instrumentations.indexOf(instr)>=0){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug("%s already initialized",moduleName);return}if(!nodule){try{nodule=require(moduleName)}catch(e){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.ERROR))logger_1.agentLogger.error("%s cannot be imported",moduleName);return}}this.initiatedInstrumentations.add(moduleName);try{instr.initialize(nodule);that.instrumentations.push(instr);if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug("%s initialized",moduleName)}catch(e){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.ERROR))logger_1.agentLogger.error("error initializing %s: %s",moduleName,e)}}minuteTicks(){try{this.reportMetrics()}catch(e){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.ERROR))logger_1.agentLogger.error(`Error reporting metrics ${e}`)}this.doOnNextFullMinute()}reportMetrics(){if(!ports_1.getPorts()){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.WARNING))logger_1.agentLogger.warning("No webapp ports detected, not sending metrics yet");return}if(this.dripStatServer.isAuthenticated()){this.sendData()}else{this.clearData();this.dripStatServer.authenticate(respType=>{})}}sendData(){this.dripStatServer.sendMetric(this._data,respType=>{if(respType==dripstat_server_1.ResponseType.OK){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug("Metrics sent")}else if(respType==dripstat_server_1.ResponseType.ERROR){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.ERROR))logger_1.agentLogger.error(`Error sending metrics ${respType}`)}else if(respType==dripstat_server_1.ResponseType.AUTH_TOKEN_EXPIRED){if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug(`Token expired, reauthenticating`);this.dripStatServer.authenticate(respType=>{if(logger_1.agentLogger.isLevelAllowed(logger_1.LogLevel.DEBUG))logger_1.agentLogger.debug(`Reauthentication resp=${respType}`)})}});this.clearData()}clearData(){this._data=new AgentData}}exports.Agent=Agent;